# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.classes import LanguageTag


class TestLanguageTag(unittest.TestCase):
    def setUp(self):
        self.language_tag = LanguageTag('a')

    def test_error(self):
        with self.assertRaises(InvalidValue):
            self.language_tag.primary_subtag = 'a1'
        with self.assertRaises(InvalidValue):
            self.language_tag.primary_subtag = ''
        with self.assertRaises(InvalidValue):
            self.language_tag.primary_subtag = 9 * 'a'

        with self.assertRaises(InvalidValue):
            self.language_tag.subsequent_subtags = ['a1', 'a#']
        with self.assertRaises(InvalidValue):
            self.language_tag.subsequent_subtags = ['a1', 9 * 'a']

        with self.assertRaises(InvalidValue):
            LanguageTag.parse_exact_size(b'')

    def test_parse(self):
        language_tag = LanguageTag.parse_exact_size(b'a')
        self.assertEqual(language_tag.primary_subtag, 'a')
        self.assertEqual(language_tag.subsequent_subtags, [])

        language_tag = LanguageTag.parse_exact_size(b'a-b-c')
        self.assertEqual(language_tag.primary_subtag, 'a')
        self.assertEqual(language_tag.subsequent_subtags, ['b', 'c', ])

    def test_compose(self):
        self.assertEqual(LanguageTag('a').compose(), b'a')
        self.assertEqual(LanguageTag('a', []).compose(), b'a')
        self.assertEqual(LanguageTag('a', ['b', 'c', ]).compose(), b'a-b-c')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import collections
import unittest

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.ike.algorithm import (
    Ikev1PayloadType, Ikev1AttributeType, Ikev1AuthenticationMethod, Ikev1DiffieHellmanGroup,
    Ikev1HashAlgorithm, Ikev1LifeType, Ikev1TransformId, Ikev1EncryptionAlgorithm,
    Ikev1Doi, Ikev1ProtocolId, Ikev1NotifyType
)

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.ike.ikev1 import (
    Ikev1AttributeAuthenticationMethod, Ikev1AttributeDiffieHellmanGroup, Ikev1AttributeHashAlgorithm,
    Ikev1AttributeLifeType, Ikev1AttributeLifeDuration, Ikev1AttributeKeyLength, Ikev1AttributeEncryptionAlgorithm,
    Ikev1PayloadTransform, Ikev1PayloadKeyExchange, Ikev1PayloadHash, Ikev1PayloadNonce, Ikev1PayloadNotification,
    Ikev1PayloadDelete, Ikev1PayloadVendorId, Ikev1PayloadCertificateRequest
)

from .classes import Ikev1PayloadBaseTest, Ikev1PayloadDoiProtocolSpiBaseTest


class TestIkev1PayloadBaseTest(unittest.TestCase):
    """Test the Ikev1PayloadBaseTest helper class, focusing only on unique aspects."""

    def setUp(self):
        self.test_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

        self.test_payload_empty = Ikev1PayloadBaseTest(test_data=b'')
        self.test_payload_with_data = Ikev1PayloadBaseTest(test_data=self.test_data)

        self.test_dict_empty = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),  # 4 bytes total (header only)
        ])
        self.test_bytes_empty = b''.join(self.test_dict_empty.values())

        self.test_dict_with_data = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 4 + 16 = 20 bytes total
            ('test_data', self.test_data),
        ])
        self.test_bytes_with_data = b''.join(self.test_dict_with_data.values())

    def test_constructor_with_test_data(self):
        payload = Ikev1PayloadBaseTest(test_data=self.test_data)
        self.assertEqual(payload.test_data, self.test_data)
        self.assertEqual(payload.next_payload, Ikev1PayloadType.NONE)

    def test_constructor_with_empty_data(self):
        payload = Ikev1PayloadBaseTest(test_data=b'')
        self.assertEqual(payload.test_data, b'')
        self.assertEqual(payload.next_payload, Ikev1PayloadType.NONE)

    def test_get_payload_type_returns_none(self):
        self.assertEqual(Ikev1PayloadBaseTest.get_payload_type(), Ikev1PayloadType.NONE)

    def test_test_data_storage(self):
        different_data = b'\xaa\xbb\xcc\xdd'
        payload = Ikev1PayloadBaseTest(test_data=different_data)
        self.assertEqual(payload.test_data, different_data)

    def test_parse_empty_test_data(self):
        parsed: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(self.test_bytes_empty)
        self.assertEqual(parsed.test_data, b'')
        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)

    def test_parse_with_test_data(self):
        parsed: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(self.test_bytes_with_data)
        self.assertEqual(parsed.test_data, self.test_data)
        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)

    def test_compose_empty_test_data(self):
        composed_bytes = self.test_payload_empty.compose()
        self.assertEqual(composed_bytes, self.test_bytes_empty)

    def test_compose_with_test_data(self):
        composed_bytes = self.test_payload_with_data.compose()
        self.assertEqual(composed_bytes, self.test_bytes_with_data)

    def test_round_trip_test_data_preservation(self):
        composed_bytes = self.test_payload_with_data.compose()
        parsed: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(composed_bytes)
        self.assertEqual(parsed.test_data, self.test_payload_with_data.test_data)


class TestIkev1PayloadBase(unittest.TestCase):
    def setUp(self):
        self.test_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

        self.test_payload_minimal = Ikev1PayloadBaseTest(
            test_data=b'',
        )
        self.test_payload_minimal.next_payload = Ikev1PayloadType.NONE

        self.test_payload_with_data = Ikev1PayloadBaseTest(
            test_data=self.test_data
        )
        self.test_payload_with_data.next_payload = Ikev1PayloadType.KEY_EXCHANGE

        self.test_dict_minimal = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),
        ])
        self.test_bytes_minimal = b''.join(self.test_dict_minimal.values())

        self.test_dict_with_data = collections.OrderedDict([
            ('next_payload', b'\x04'),  # KEY_EXCHANGE = 0x04
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 4 + 16 = 20 bytes total
            ('test_data', self.test_data),
        ])
        self.test_bytes_with_data = b''.join(self.test_dict_with_data.values())

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadBaseTest.get_payload_type(), Ikev1PayloadType.NONE)

    def test_parse(self):
        parsed_minimal: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(self.test_bytes_minimal)
        self.assertEqual(parsed_minimal.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed_minimal.test_data, b'')

        parsed_with_data: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(self.test_bytes_with_data)
        self.assertEqual(parsed_with_data.next_payload, Ikev1PayloadType.KEY_EXCHANGE)
        self.assertEqual(parsed_with_data.test_data, self.test_data)

    def test_compose(self):
        composed_minimal = self.test_payload_minimal.compose()
        self.assertEqual(composed_minimal, self.test_bytes_minimal)

        composed_with_data = self.test_payload_with_data.compose()
        self.assertEqual(composed_with_data, self.test_bytes_with_data)

    def test_round_trip(self):
        composed_minimal = self.test_payload_minimal.compose()
        parsed_minimal: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(composed_minimal)

        self.assertEqual(parsed_minimal.test_data, self.test_payload_minimal.test_data)
        self.assertEqual(parsed_minimal.next_payload, self.test_payload_minimal.next_payload)

        composed_with_data = self.test_payload_with_data.compose()
        parsed_with_data: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(composed_with_data)

        self.assertEqual(parsed_with_data.test_data, self.test_payload_with_data.test_data)
        self.assertEqual(parsed_with_data.next_payload, self.test_payload_with_data.next_payload)

    def test_next_payload(self):
        self.test_payload_minimal.next_payload = Ikev1PayloadType.SECURITY_ASSOCIATION
        composed = self.test_payload_minimal.compose()
        parsed: Ikev1PayloadBaseTest = Ikev1PayloadBaseTest.parse_exact_size(composed)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.SECURITY_ASSOCIATION)

    def test_payload_length(self):
        test_data = b'\x00\x01\x02\x03\x04\x05'
        payload = Ikev1PayloadBaseTest(test_data)
        payload.next_payload = Ikev1PayloadType.NONCE

        composed = payload.compose()
        self.assertEqual(len(composed), 10)  # 1 + 1 + 2 + 6 = 10 bytes total

        self.assertEqual(composed[2:4], b'\x00\x0a')  # 10 = 0x000a

    def test_error_parse_not_enough_data(self):
        incomplete_data = self.test_bytes_minimal[:-1]

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadBaseTest.parse_exact_size(incomplete_data)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_parse_payload_length_mismatch(self):
        malformed_dict = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 20 bytes total
        ])
        malformed_data = b''.join(malformed_dict.values())

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadBaseTest.parse_exact_size(malformed_data)

        self.assertEqual(context_manager.exception.bytes_needed, 16)


class TestIkev1PayloadSaIdentifyingBase(unittest.TestCase):
    """Test Ikev1PayloadDoiProtocolSpiBase (RFC 2408 Section 2.4 Identifying Security Associations)."""

    def setUp(self):
        self.doi = Ikev1Doi.IPSEC
        self.protocol_id = Ikev1ProtocolId.ISAKMP
        self.spi_size = 8
        self.extra_data = b'\x00\x01\x02\x03'

        self.test_payload = Ikev1PayloadDoiProtocolSpiBaseTest(
            doi=self.doi,
            protocol_id=self.protocol_id,
            spi_size=self.spi_size,
            extra_data=self.extra_data
        )
        self.test_payload.next_payload = Ikev1PayloadType.NONE

    def test_parse(self):
        composed = self.test_payload.compose()
        parsed: Ikev1PayloadDoiProtocolSpiBaseTest = Ikev1PayloadDoiProtocolSpiBaseTest.parse_exact_size(composed)
        self.assertEqual(parsed.doi, self.doi)
        self.assertEqual(parsed.protocol_id, self.protocol_id)
        self.assertEqual(parsed.spi_size, self.spi_size)
        self.assertEqual(parsed.extra_data, self.extra_data)

    def test_compose_round_trip(self):
        composed = self.test_payload.compose()
        parsed: Ikev1PayloadDoiProtocolSpiBaseTest = Ikev1PayloadDoiProtocolSpiBaseTest.parse_exact_size(composed)
        recomposed = parsed.compose()
        self.assertEqual(recomposed, composed)

    def test_different_doi_protocol_spi_values(self):
        for doi, protocol_id, spi_size in [
            (Ikev1Doi.ISAKMP, Ikev1ProtocolId.IPSEC_ESP, 4),
            (Ikev1Doi.GDOI, Ikev1ProtocolId.IPSEC_AH, 0),
        ]:
            payload = Ikev1PayloadDoiProtocolSpiBaseTest(
                doi=doi,
                protocol_id=protocol_id,
                spi_size=spi_size,
                extra_data=b''
            )
            composed = payload.compose()
            parsed: Ikev1PayloadDoiProtocolSpiBaseTest = Ikev1PayloadDoiProtocolSpiBaseTest.parse_exact_size(composed)
            self.assertEqual(parsed.doi, doi)
            self.assertEqual(parsed.protocol_id, protocol_id)
            self.assertEqual(parsed.spi_size, spi_size)

    def test_error_not_enough_data(self):
        minimal = Ikev1PayloadDoiProtocolSpiBaseTest(
            doi=self.doi,
            protocol_id=self.protocol_id,
            spi_size=0,
            extra_data=b''
        )
        minimal.next_payload = Ikev1PayloadType.NONE

        minimal_bytes = minimal.compose()
        incomplete = minimal_bytes[:-1]

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadDoiProtocolSpiBaseTest.parse_exact_size(incomplete)
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        truncated_doi_protocol_spi = (
            b'\x00\x00'  # next_payload, reserved
            b'\x00\x09'  # payload_length = 9
            b'\x00\x00\x00\x00\x00'  # 5 bytes (need 6 for DOI+Protocol+SPI)
        )
        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadDoiProtocolSpiBaseTest.parse_exact_size(truncated_doi_protocol_spi)
        self.assertEqual(context_manager.exception.bytes_needed, 1)


class TestIkev1AttributeKeyLength(unittest.TestCase):
    def setUp(self):
        self.key_length_value = 128  # 128-bit key length
        self.key_length_attribute = Ikev1AttributeKeyLength(value=self.key_length_value)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeKeyLength.get_type(), Ikev1AttributeType.KEY_LENGTH)

    def test_key_length_value_support(self):
        different_key_lengths = [64, 128, 192, 256]  # bits

        for key_length in different_key_lengths:
            attribute = Ikev1AttributeKeyLength(value=key_length)
            self.assertEqual(attribute.value, key_length)

    def test_round_trip(self):
        composed_bytes = self.key_length_attribute.compose()
        parsed_attribute: Ikev1AttributeKeyLength = Ikev1AttributeKeyLength.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_attribute.value, self.key_length_attribute.value)


class TestIkev1AttributeEncryptionAlgorithm(unittest.TestCase):
    def setUp(self):
        self.encryption_algorithm = Ikev1EncryptionAlgorithm.AES_CBC
        self.encryption_attribute = Ikev1AttributeEncryptionAlgorithm(value=self.encryption_algorithm)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeEncryptionAlgorithm.get_type(), Ikev1AttributeType.ENCRYPTION_ALGORITHM)

    def test_get_enum_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeEncryptionAlgorithm._get_enum_type(), Ikev1EncryptionAlgorithm)

    def test_encryption_algorithm_support(self):
        different_encryption_algorithms = [
            Ikev1EncryptionAlgorithm.DES_CBC,
            Ikev1EncryptionAlgorithm.DES3_CBC,
            Ikev1EncryptionAlgorithm.AES_CBC,
        ]

        for encryption_algorithm in different_encryption_algorithms:
            attribute = Ikev1AttributeEncryptionAlgorithm(value=encryption_algorithm)
            self.assertEqual(attribute.value, encryption_algorithm)

    def test_round_trip(self):
        composed_bytes = self.encryption_attribute.compose()
        parsed_attribute: Ikev1AttributeEncryptionAlgorithm = Ikev1AttributeEncryptionAlgorithm.parse_exact_size(
            composed_bytes
        )

        self.assertEqual(parsed_attribute.value, self.encryption_attribute.value)


class TestIkev1AttributeAuthenticationMethod(unittest.TestCase):
    def setUp(self):
        self.auth_method = Ikev1AuthenticationMethod.PRE_SHARED_KEY
        self.auth_attribute = Ikev1AttributeAuthenticationMethod(value=self.auth_method)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeAuthenticationMethod.get_type(), Ikev1AttributeType.AUTHENTICATION_METHOD)

    def test_get_enum_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeAuthenticationMethod._get_enum_type(), Ikev1AuthenticationMethod)

    def test_authentication_method_support(self):
        different_auth_methods = [
            Ikev1AuthenticationMethod.PRE_SHARED_KEY,
            Ikev1AuthenticationMethod.DSS_SIGNATURES,
            Ikev1AuthenticationMethod.RSA_SIGNATURES,
        ]

        for auth_method in different_auth_methods:
            attribute = Ikev1AttributeAuthenticationMethod(value=auth_method)
            self.assertEqual(attribute.value, auth_method)

    def test_round_trip(self):
        composed_bytes = self.auth_attribute.compose()
        parsed_attribute: Ikev1AttributeAuthenticationMethod = Ikev1AttributeAuthenticationMethod.parse_exact_size(
            composed_bytes
        )

        self.assertEqual(parsed_attribute.value, self.auth_attribute.value)

    def test_error_invalid_value(self):
        with self.assertRaises(InvalidValue):
            Ikev1AttributeAuthenticationMethod(value="invalid")


class TestIkev1AttributeDiffieHellmanGroup(unittest.TestCase):
    def setUp(self):
        self.dh_group = Ikev1DiffieHellmanGroup.MODP_1024_BIT
        self.dh_attribute = Ikev1AttributeDiffieHellmanGroup(value=self.dh_group)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeDiffieHellmanGroup.get_type(), Ikev1AttributeType.GROUP_DESCRIPTION)

    def test_get_enum_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeDiffieHellmanGroup._get_enum_type(), Ikev1DiffieHellmanGroup)

    def test_diffie_hellman_group_support(self):
        different_dh_groups = [
            Ikev1DiffieHellmanGroup.MODP_768_BIT,
            Ikev1DiffieHellmanGroup.MODP_1024_BIT,
            Ikev1DiffieHellmanGroup.MODP_1536_BIT,
        ]

        for dh_group in different_dh_groups:
            attribute = Ikev1AttributeDiffieHellmanGroup(value=dh_group)
            self.assertEqual(attribute.value, dh_group)

    def test_round_trip(self):
        composed_bytes = self.dh_attribute.compose()
        parsed_attribute: Ikev1AttributeDiffieHellmanGroup = Ikev1AttributeDiffieHellmanGroup.parse_exact_size(
            composed_bytes
        )

        self.assertEqual(parsed_attribute.value, self.dh_attribute.value)


class TestIkev1AttributeHashAlgorithm(unittest.TestCase):
    def setUp(self):
        self.hash_algorithm = Ikev1HashAlgorithm.MD5
        self.hash_attribute = Ikev1AttributeHashAlgorithm(value=self.hash_algorithm)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeHashAlgorithm.get_type(), Ikev1AttributeType.HASH_ALGORITHM)

    def test_get_enum_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeHashAlgorithm._get_enum_type(), Ikev1HashAlgorithm)

    def test_hash_algorithm_support(self):
        different_hash_algorithms = [
            Ikev1HashAlgorithm.MD5,
            Ikev1HashAlgorithm.SHA,
        ]

        for hash_algorithm in different_hash_algorithms:
            attribute = Ikev1AttributeHashAlgorithm(value=hash_algorithm)
            self.assertEqual(attribute.value, hash_algorithm)

    def test_round_trip(self):
        composed_bytes = self.hash_attribute.compose()
        parsed_attribute: Ikev1AttributeHashAlgorithm = Ikev1AttributeHashAlgorithm.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_attribute.value, self.hash_attribute.value)


class TestIkev1AttributeLifeType(unittest.TestCase):
    def setUp(self):
        self.life_type = Ikev1LifeType.SECONDS
        self.life_type_attribute = Ikev1AttributeLifeType(value=self.life_type)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeLifeType.get_type(), Ikev1AttributeType.LIFE_TYPE)

    def test_get_enum_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeLifeType._get_enum_type(), Ikev1LifeType)

    def test_life_type_support(self):
        different_life_types = [
            Ikev1LifeType.SECONDS,
            Ikev1LifeType.KILOBYTES,
        ]

        for life_type in different_life_types:
            attribute = Ikev1AttributeLifeType(value=life_type)
            self.assertEqual(attribute.value, life_type)

    def test_round_trip(self):
        composed_bytes = self.life_type_attribute.compose()
        parsed_attribute: Ikev1AttributeLifeType = Ikev1AttributeLifeType.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_attribute.value, self.life_type_attribute.value)


class TestIkev1AttributeLifeDuration(unittest.TestCase):
    def setUp(self):
        self.life_duration_value = 3600  # 1 hour in seconds
        self.life_duration_attribute = Ikev1AttributeLifeDuration(value=self.life_duration_value)

    def test_get_type(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeLifeDuration.get_type(), Ikev1AttributeType.LIFE_DURATION)

    def test_get_size(self):
        # pylint: disable=protected-access
        self.assertEqual(Ikev1AttributeLifeDuration._get_size(), 4)

    def test_life_duration_value_support(self):
        different_durations = [3600, 86400, 604800]  # seconds: 1 hour, 1 day, 1 week

        for duration in different_durations:
            attribute = Ikev1AttributeLifeDuration(value=duration)
            self.assertEqual(attribute.value, duration)

    def test_round_trip(self):
        composed_bytes = self.life_duration_attribute.compose()
        parsed_attribute: Ikev1AttributeLifeDuration = Ikev1AttributeLifeDuration.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_attribute.value, self.life_duration_attribute.value)


class TestIkev1PayloadTransform(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.transform_id = Ikev1TransformId.KEY_IKE
        self.attributes = []
        self.transform = Ikev1PayloadTransform(transform_id=self.transform_id, attributes=self.attributes)

        # Test data for transform without attributes
        self.test_dict_no_attrs = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x08'),
            ('transform_number', b'\x01'),  # Transform number 1
            ('transform_id', b'\x01'),  # KEY_IKE = 0x01
            ('reserved2', b'\x00\x00'),
        ])
        self.test_bytes_no_attrs = b''.join(self.test_dict_no_attrs.values())

        # Test data for transform with authentication method attribute
        self.auth_attribute = Ikev1AttributeAuthenticationMethod(value=Ikev1AuthenticationMethod.PRE_SHARED_KEY)
        self.test_dict_auth_attr = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x0c'),  # 12 bytes total
            ('transform_number', b'\x01'),  # Transform number 1
            ('transform_id', b'\x01'),  # KEY_IKE = 0x01
            ('reserved2', b'\x00\x00'),
            ('attr_format_type', b'\x80\x03'),  # AF=1, type=3 (AUTH_METHOD)
            ('attr_value', b'\x00\x01'),  # PSK = 0x0001
        ])
        self.test_bytes_auth_attr = b''.join(self.test_dict_auth_attr.values())

        # Test data for transform with key length attribute
        self.key_length_attribute = Ikev1AttributeKeyLength(value=128)
        self.test_dict_key_length_attr = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x0c'),  # 12 bytes total
            ('transform_number', b'\x01'),  # Transform number 1
            ('transform_id', b'\x01'),  # KEY_IKE = 0x01
            ('reserved2', b'\x00\x00'),
            ('attr_format_type', b'\x80\x0e'),  # AF=1, type=14 (KEY_LENGTH)
            ('attr_value', b'\x00\x80'),  # 128 = 0x0080
        ])
        self.test_bytes_key_length_attr = b''.join(self.test_dict_key_length_attr.values())

        # Test data for transform with life duration attribute
        self.life_duration_attribute = Ikev1AttributeLifeDuration(value=3600)
        self.test_dict_life_duration_attr = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x10'),  # 16 bytes total
            ('transform_number', b'\x01'),  # Transform number 1
            ('transform_id', b'\x01'),  # KEY_IKE = 0x01
            ('reserved2', b'\x00\x00'),
            ('attr_format_type', b'\x00\x0c'),  # AF=0, type=12 (LIFE_DURATION)
            ('attr_length', b'\x00\x04'),  # 4 bytes value length
            ('attr_value', b'\x00\x00\x0e\x10'),  # 3600 = 0x00000e10
        ])
        self.test_bytes_life_duration_attr = b''.join(self.test_dict_life_duration_attr.values())

        # Test data for transform with multiple attributes
        self.test_dict_multi_attrs = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x10'),  # 16 bytes total
            ('transform_number', b'\x01'),  # Transform number 1
            ('transform_id', b'\x01'),  # KEY_IKE = 0x01
            ('reserved2', b'\x00\x00'),
            ('attr1_format_type', b'\x80\x03'),  # AF=1, type=3 (AUTH_METHOD)
            ('attr1_value', b'\x00\x01'),  # PSK = 0x0001
            ('attr2_format_type', b'\x80\x0e'),  # AF=1, type=14 (KEY_LENGTH)
            ('attr2_value', b'\x00\x80'),  # 128 = 0x0080
        ])
        self.test_bytes_multi_attrs = b''.join(self.test_dict_multi_attrs.values())

        # Transform objects for compose tests
        self.transform_auth_attr = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.auth_attribute]
        )
        self.transform_auth_attr.transform_number = 1
        self.transform_auth_attr.next_payload = Ikev1PayloadType.NONE

        self.transform_key_length_attr = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.key_length_attribute]
        )
        self.transform_key_length_attr.transform_number = 1
        self.transform_key_length_attr.next_payload = Ikev1PayloadType.NONE

        self.transform_life_duration_attr = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.life_duration_attribute]
        )
        self.transform_life_duration_attr.transform_number = 1
        self.transform_life_duration_attr.next_payload = Ikev1PayloadType.NONE

        self.transform_multi_attrs = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.auth_attribute, self.key_length_attribute]
        )
        self.transform_multi_attrs.transform_number = 1
        self.transform_multi_attrs.next_payload = Ikev1PayloadType.NONE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadTransform.get_payload_type(), Ikev1PayloadType.TRANSFORM)

    def test_transform_id_storage(self):
        self.assertEqual(self.transform.transform_id, self.transform_id)

    def test_attributes_storage(self):
        self.assertEqual(self.transform.attributes, self.attributes)

    def test_get_attribute_by_type(self):
        transform = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[self.auth_attribute],
        )

        attribute = transform.get_attribute_by_type(Ikev1AttributeType.AUTHENTICATION_METHOD)
        self.assertIsInstance(attribute, Ikev1AttributeAuthenticationMethod)
        self.assertEqual(attribute.value, Ikev1AuthenticationMethod.PRE_SHARED_KEY)

    def test_error_get_attribute_by_type_not_found(self):
        transform = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[],
        )

        with self.assertRaises(KeyError):
            transform.get_attribute_by_type(Ikev1AttributeType.AUTHENTICATION_METHOD)

    def test_transform_number_initialization(self):
        self.assertIsNone(self.transform.transform_number)

    def test_transform_with_attributes(self):
        auth_attribute = Ikev1AttributeAuthenticationMethod(value=Ikev1AuthenticationMethod.PRE_SHARED_KEY)
        transform_with_attrs = Ikev1PayloadTransform(
            transform_id=Ikev1TransformId.KEY_IKE,
            attributes=[auth_attribute]
        )

        self.assertEqual(len(transform_with_attrs.attributes), 1)
        self.assertEqual(transform_with_attrs.attributes[0], auth_attribute)

    def test_parse(self):
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(self.test_bytes_no_attrs)

        self.assertEqual(parsed_transform.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed_transform.transform_id, Ikev1TransformId.KEY_IKE)
        self.assertEqual(len(parsed_transform.attributes), 0)

    def test_compose(self):
        self.transform.transform_number = 1
        self.transform.next_payload = Ikev1PayloadType.NONE

        composed_bytes = self.transform.compose()
        self.assertEqual(composed_bytes, self.test_bytes_no_attrs)

    def test_round_trip(self):
        self.transform.transform_number = 1
        self.transform.next_payload = Ikev1PayloadType.NONE

        composed_bytes = self.transform.compose()
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_transform.transform_id, self.transform.transform_id)
        self.assertEqual(parsed_transform.next_payload, self.transform.next_payload)
        self.assertEqual(len(parsed_transform.attributes), len(self.transform.attributes))

    def test_parse_with_attributes(self):
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(self.test_bytes_auth_attr)

        self.assertEqual(parsed_transform.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed_transform.transform_id, Ikev1TransformId.KEY_IKE)
        self.assertEqual(len(parsed_transform.attributes), 1)

        attribute = parsed_transform.attributes[0]
        self.assertIsInstance(attribute, Ikev1AttributeAuthenticationMethod)
        self.assertEqual(attribute.value, Ikev1AuthenticationMethod.PRE_SHARED_KEY)

    def test_parse_with_key_length_attribute(self):
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(
            self.test_bytes_key_length_attr
        )

        self.assertEqual(len(parsed_transform.attributes), 1)
        attribute = parsed_transform.attributes[0]
        self.assertIsInstance(attribute, Ikev1AttributeKeyLength)
        self.assertEqual(attribute.value, 128)

    def test_parse_with_life_duration_attribute(self):
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(
            self.test_bytes_life_duration_attr
        )

        self.assertEqual(len(parsed_transform.attributes), 1)
        attribute = parsed_transform.attributes[0]
        self.assertIsInstance(attribute, Ikev1AttributeLifeDuration)
        self.assertEqual(attribute.value, 3600)

    def test_compose_with_attributes(self):
        composed_bytes = self.transform_auth_attr.compose()
        self.assertEqual(composed_bytes, self.test_bytes_auth_attr)

    def test_compose_with_key_length_attribute(self):
        composed_bytes = self.transform_key_length_attr.compose()
        self.assertEqual(composed_bytes, self.test_bytes_key_length_attr)

    def test_compose_with_life_duration_attribute(self):
        composed_bytes = self.transform_life_duration_attr.compose()
        self.assertEqual(composed_bytes, self.test_bytes_life_duration_attr)

    def test_compose_with_multiple_attributes(self):
        composed_bytes = self.transform_multi_attrs.compose()
        self.assertEqual(composed_bytes, self.test_bytes_multi_attrs)

    def test_round_trip_with_attributes(self):
        composed_bytes = self.transform_multi_attrs.compose()
        parsed_transform: Ikev1PayloadTransform = Ikev1PayloadTransform.parse_exact_size(composed_bytes)

        self.assertEqual(parsed_transform.transform_id, self.transform_multi_attrs.transform_id)
        self.assertEqual(parsed_transform.next_payload, self.transform_multi_attrs.next_payload)
        self.assertEqual(len(parsed_transform.attributes), len(self.transform_multi_attrs.attributes))

        # Verify individual attributes
        self.assertIsInstance(parsed_transform.attributes[0], Ikev1AttributeAuthenticationMethod)
        self.assertEqual(parsed_transform.attributes[0].value, Ikev1AuthenticationMethod.PRE_SHARED_KEY)
        self.assertIsInstance(parsed_transform.attributes[1], Ikev1AttributeKeyLength)
        self.assertEqual(parsed_transform.attributes[1].value, 128)


class TestIkev1PayloadKeyExchange(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.key_exchange_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        self.large_key_data = bytes(range(256))

        self.test_dict_small_key = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 4 + 16 = 20 bytes total
            ('key_exchange_data', self.key_exchange_data),
        ])
        self.test_bytes_small_key = b''.join(self.test_dict_small_key.values())

        self.test_dict_empty_key = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),
        ])
        self.test_bytes_empty_key = b''.join(self.test_dict_empty_key.values())

        self.test_dict_large_key = collections.OrderedDict([
            ('next_payload', b'\x0a'),  # NONCE = 0x0a (10)
            ('reserved', b'\x00'),
            ('payload_length', b'\x01\x04'),  # 4 + 256 = 260 bytes total
            ('key_exchange_data', self.large_key_data),
        ])
        self.test_bytes_large_key = b''.join(self.test_dict_large_key.values())

        # KE objects for compose tests
        self.ke_small = Ikev1PayloadKeyExchange(key_exchange_data=self.key_exchange_data)
        self.ke_small.next_payload = Ikev1PayloadType.NONE

        self.ke_empty = Ikev1PayloadKeyExchange(key_exchange_data=b'')
        self.ke_empty.next_payload = Ikev1PayloadType.NONE

        self.ke_large = Ikev1PayloadKeyExchange(key_exchange_data=self.large_key_data)
        self.ke_large.next_payload = Ikev1PayloadType.NONCE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadKeyExchange.get_payload_type(), Ikev1PayloadType.KEY_EXCHANGE)

    def test_constructor_with_key_data(self):
        ke = Ikev1PayloadKeyExchange(key_exchange_data=self.key_exchange_data)
        self.assertEqual(ke.key_exchange_data, self.key_exchange_data)
        self.assertEqual(ke.next_payload, Ikev1PayloadType.NONE)

    def test_constructor_with_empty_data(self):
        ke = Ikev1PayloadKeyExchange(key_exchange_data=b'')
        self.assertEqual(ke.key_exchange_data, b'')

    def test_constructor_with_large_data(self):
        ke = Ikev1PayloadKeyExchange(key_exchange_data=self.large_key_data)
        self.assertEqual(ke.key_exchange_data, self.large_key_data)
        self.assertEqual(len(ke.key_exchange_data), 256)

    def test_key_exchange_data_storage(self):
        different_data = b'\xaa\xbb\xcc\xdd\xee\xff'
        ke = Ikev1PayloadKeyExchange(key_exchange_data=different_data)
        self.assertEqual(ke.key_exchange_data, different_data)

    def test_parse_small_key_data(self):
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(self.test_bytes_small_key)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed.key_exchange_data, self.key_exchange_data)

    def test_parse_empty_key_data(self):
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(self.test_bytes_empty_key)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed.key_exchange_data, b'')

    def test_parse_large_key_data(self):
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(self.test_bytes_large_key)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONCE)
        self.assertEqual(parsed.key_exchange_data, self.large_key_data)
        self.assertEqual(len(parsed.key_exchange_data), 256)

    def test_compose_small_key_data(self):
        composed_bytes = self.ke_small.compose()
        self.assertEqual(composed_bytes, self.test_bytes_small_key)

    def test_compose_empty_key_data(self):
        composed_bytes = self.ke_empty.compose()
        self.assertEqual(composed_bytes, self.test_bytes_empty_key)

    def test_compose_large_key_data(self):
        composed_bytes = self.ke_large.compose()
        self.assertEqual(composed_bytes, self.test_bytes_large_key)

    def test_round_trip_small_key(self):
        composed_bytes = self.ke_small.compose()
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.key_exchange_data, self.ke_small.key_exchange_data)
        self.assertEqual(parsed.next_payload, self.ke_small.next_payload)

    def test_round_trip_empty_key(self):
        composed_bytes = self.ke_empty.compose()
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.key_exchange_data, self.ke_empty.key_exchange_data)
        self.assertEqual(parsed.next_payload, self.ke_empty.next_payload)

    def test_round_trip_large_key(self):
        composed_bytes = self.ke_large.compose()
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.key_exchange_data, self.ke_large.key_exchange_data)
        self.assertEqual(parsed.next_payload, self.ke_large.next_payload)
        self.assertEqual(len(parsed.key_exchange_data), len(self.ke_large.key_exchange_data))

    def test_payload_length_calculation(self):
        test_data = b'\x12\x34\x56\x78\x9a\xbc'
        ke = Ikev1PayloadKeyExchange(key_exchange_data=test_data)
        ke.next_payload = Ikev1PayloadType.VENDOR_ID

        composed = ke.compose()
        self.assertEqual(len(composed), 10)  # 4 + 6 = 10 bytes total
        self.assertEqual(composed[2:4], b'\x00\x0a')  # 10 = 0x000a

    def test_different_next_payload_types(self):
        ke = Ikev1PayloadKeyExchange(key_exchange_data=self.key_exchange_data)
        ke.next_payload = Ikev1PayloadType.SECURITY_ASSOCIATION

        composed = ke.compose()
        parsed: Ikev1PayloadKeyExchange = Ikev1PayloadKeyExchange.parse_exact_size(composed)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.SECURITY_ASSOCIATION)
        self.assertEqual(parsed.key_exchange_data, self.key_exchange_data)


class TestIkev1PayloadHash(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.hash_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        self.large_hash_data = bytes(range(256))

        self.test_dict_small_hash = collections.OrderedDict([
            ('next_payload', b'\x0a'),  # NONCE = 0x0a (10)
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 4 + 16 = 20 bytes total
            ('hash_data', self.hash_data),
        ])
        self.test_bytes_small_hash = b''.join(self.test_dict_small_hash.values())

        self.test_dict_empty_hash = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),
        ])
        self.test_bytes_empty_hash = b''.join(self.test_dict_empty_hash.values())

        self.test_dict_large_hash = collections.OrderedDict([
            ('next_payload', b'\x04'),  # KEY_EXCHANGE = 0x04
            ('reserved', b'\x00'),
            ('payload_length', b'\x01\x04'),  # 4 + 256 = 260 bytes total
            ('hash_data', self.large_hash_data),
        ])
        self.test_bytes_large_hash = b''.join(self.test_dict_large_hash.values())

        self.hash_small = Ikev1PayloadHash(hash_data=self.hash_data)
        self.hash_small.next_payload = Ikev1PayloadType.NONCE

        self.hash_empty = Ikev1PayloadHash(hash_data=b'')
        self.hash_empty.next_payload = Ikev1PayloadType.NONE

        self.hash_large = Ikev1PayloadHash(hash_data=self.large_hash_data)
        self.hash_large.next_payload = Ikev1PayloadType.KEY_EXCHANGE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadHash.get_payload_type(), Ikev1PayloadType.HASH)

    def test_parse_small_hash_data(self):
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(self.test_bytes_small_hash)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONCE)
        self.assertEqual(parsed.hash_data, self.hash_data)

    def test_parse_empty_hash_data(self):
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(self.test_bytes_empty_hash)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed.hash_data, b'')

    def test_parse_large_hash_data(self):
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(self.test_bytes_large_hash)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.KEY_EXCHANGE)
        self.assertEqual(parsed.hash_data, self.large_hash_data)
        self.assertEqual(len(parsed.hash_data), 256)

    def test_compose_small_hash_data(self):
        composed_bytes = self.hash_small.compose()
        self.assertEqual(composed_bytes, self.test_bytes_small_hash)

    def test_compose_empty_hash_data(self):
        composed_bytes = self.hash_empty.compose()
        self.assertEqual(composed_bytes, self.test_bytes_empty_hash)

    def test_compose_large_hash_data(self):
        composed_bytes = self.hash_large.compose()
        self.assertEqual(composed_bytes, self.test_bytes_large_hash)

    def test_round_trip_small_hash(self):
        composed_bytes = self.hash_small.compose()
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.hash_data, self.hash_small.hash_data)
        self.assertEqual(parsed.next_payload, self.hash_small.next_payload)

    def test_round_trip_empty_hash(self):
        composed_bytes = self.hash_empty.compose()
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.hash_data, self.hash_empty.hash_data)
        self.assertEqual(parsed.next_payload, self.hash_empty.next_payload)

    def test_round_trip_large_hash(self):
        composed_bytes = self.hash_large.compose()
        parsed: Ikev1PayloadHash = Ikev1PayloadHash.parse_exact_size(composed_bytes)

        self.assertEqual(parsed.hash_data, self.hash_large.hash_data)
        self.assertEqual(parsed.next_payload, self.hash_large.next_payload)
        self.assertEqual(len(parsed.hash_data), len(self.hash_large.hash_data))

    def test_payload_length_calculation(self):
        test_data = b'\x00\x01\x02\x03\x04\x05'
        payload = Ikev1PayloadHash(hash_data=test_data)
        payload.next_payload = Ikev1PayloadType.VENDOR_ID

        composed = payload.compose()
        self.assertEqual(len(composed), 10)  # 4 + 6 = 10 bytes total
        self.assertEqual(composed[2:4], b'\x00\x0a')  # 10 = 0x000a

    def test_error_parse_not_enough_data(self):
        incomplete_data = b'\x00\x00\x00'

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadHash.parse_exact_size(incomplete_data)

        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_error_parse_payload_length_mismatch(self):
        malformed_dict = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),  # 20 bytes total, but missing data
        ])
        malformed_data = b''.join(malformed_dict.values())

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadHash.parse_exact_size(malformed_data)

        self.assertEqual(context_manager.exception.bytes_needed, 16)


class TestIkev1PayloadNonce(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.nonce_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        self.large_nonce_data = bytes(range(256))

        self.test_dict_small_nonce = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x14'),
            ('nonce_data', self.nonce_data),
        ])
        self.test_bytes_small_nonce = b''.join(self.test_dict_small_nonce.values())

        self.test_dict_empty_nonce = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),
        ])
        self.test_bytes_empty_nonce = b''.join(self.test_dict_empty_nonce.values())

        self.test_dict_large_nonce = collections.OrderedDict([
            ('next_payload', b'\x04'),  # KEY_EXCHANGE = 0x04
            ('reserved', b'\x00'),
            ('payload_length', b'\x01\x04'),
            ('nonce_data', self.large_nonce_data),
        ])
        self.test_bytes_large_nonce = b''.join(self.test_dict_large_nonce.values())

        self.nonce_small = Ikev1PayloadNonce(nonce_data=self.nonce_data)
        self.nonce_small.next_payload = Ikev1PayloadType.NONE

        self.nonce_empty = Ikev1PayloadNonce(nonce_data=b'')
        self.nonce_empty.next_payload = Ikev1PayloadType.NONE

        self.nonce_large = Ikev1PayloadNonce(nonce_data=self.large_nonce_data)
        self.nonce_large.next_payload = Ikev1PayloadType.KEY_EXCHANGE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadNonce.get_payload_type(), Ikev1PayloadType.NONCE)

    def test_payload_length_calculation(self):
        test_data = b'\x12\x34\x56\x78\x9a\xbc'
        nonce = Ikev1PayloadNonce(nonce_data=test_data)
        nonce.next_payload = Ikev1PayloadType.VENDOR_ID

        composed = nonce.compose()
        self.assertEqual(len(composed), 10)
        self.assertEqual(composed[2:4], b'\x00\x0a')  # 10 = 0x000a

    def test_different_next_payload_types(self):
        nonce = Ikev1PayloadNonce(nonce_data=self.nonce_data)
        nonce.next_payload = Ikev1PayloadType.SECURITY_ASSOCIATION

        composed = nonce.compose()
        parsed: Ikev1PayloadNonce = Ikev1PayloadNonce.parse_exact_size(composed)

        self.assertEqual(parsed.next_payload, Ikev1PayloadType.SECURITY_ASSOCIATION)
        self.assertEqual(parsed.nonce_data, self.nonce_data)

    def test_nonce_with_special_characters(self):
        special_nonce = b'Nonce\x00\x01\x02\x03\xff\xfe\xfd'
        nonce = Ikev1PayloadNonce(nonce_data=special_nonce)
        nonce.next_payload = Ikev1PayloadType.NOTIFICATION

        composed = nonce.compose()
        parsed: Ikev1PayloadNonce = Ikev1PayloadNonce.parse_exact_size(composed)

        self.assertEqual(parsed.nonce_data, special_nonce)
        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NOTIFICATION)
        self.assertEqual(len(parsed.nonce_data), len(special_nonce))


class TestIkev1PayloadNotification(unittest.TestCase):
    """Test Notification-specific payload fields: notify_type, spi, notification_data."""

    def setUp(self):
        self.notify_type = Ikev1NotifyType.NO_PROPOSAL_CHOSEN
        self.spi = b'\x00\x01\x02\x03\x04\x05\x06\x07'
        self.notification_data = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadNotification.get_payload_type(), Ikev1PayloadType.NOTIFICATION)

    def test_payload_length_calculation(self):
        test_spi = b'\x12\x34\x56\x78'
        test_notification_data = b'\x9a\xbc\xde\xf0'
        notification = Ikev1PayloadNotification(
            doi=Ikev1Doi.IPSEC,
            protocol_id=Ikev1ProtocolId.IPSEC_ESP,
            spi_size=4,
            notify_type=Ikev1NotifyType.INVALID_SPI,
            spi=test_spi,
            notification_data=test_notification_data
        )
        notification.next_payload = Ikev1PayloadType.VENDOR_ID

        composed = notification.compose()
        self.assertEqual(len(composed), 20)
        self.assertEqual(composed[2:4], b'\x00\x14')  # 20 = 0x0014

    def test_notify_type_spi_notification_data_round_trip(self):
        notification = Ikev1PayloadNotification(
            doi=Ikev1Doi.IPSEC,
            protocol_id=Ikev1ProtocolId.ISAKMP,
            spi_size=8,
            notify_type=self.notify_type,
            spi=self.spi,
            notification_data=self.notification_data
        )
        composed = notification.compose()
        parsed: Ikev1PayloadNotification = Ikev1PayloadNotification.parse_exact_size(composed)
        self.assertEqual(parsed.notify_type, self.notify_type)
        self.assertEqual(parsed.spi, self.spi)
        self.assertEqual(parsed.notification_data, self.notification_data)


class TestIkev1PayloadDelete(unittest.TestCase):
    def setUp(self):
        self.doi = Ikev1Doi.IPSEC
        self.protocol_id = Ikev1ProtocolId.ISAKMP
        self.spi_size = 8
        self.spis = [
            b'\x00\x01\x02\x03\x04\x05\x06\x07',
            b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
        ]

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadDelete.get_payload_type(), Ikev1PayloadType.DELETE)

    def test_payload_length_calculation(self):
        spis_4byte = [b'\x12\x34\x56\x78', b'\x9a\xbc\xde\xf0']
        delete = Ikev1PayloadDelete(
            doi=Ikev1Doi.IPSEC,
            protocol_id=Ikev1ProtocolId.IPSEC_ESP,
            spi_size=4,
            spis=spis_4byte,
        )
        delete.next_payload = Ikev1PayloadType.VENDOR_ID

        composed = delete.compose()
        self.assertEqual(len(composed), 20)  # 4 + 4 + 1 + 1 + 2 + 8 = 20 bytes total
        self.assertEqual(composed[2:4], b'\x00\x14')  # 20 = 0x0014

    def test_spis_round_trip(self):
        delete = Ikev1PayloadDelete(
            doi=self.doi,
            protocol_id=self.protocol_id,
            spi_size=self.spi_size,
            spis=self.spis,
        )
        composed = delete.compose()
        parsed: Ikev1PayloadDelete = Ikev1PayloadDelete.parse_exact_size(composed)
        self.assertEqual(parsed.spis, self.spis)


class TestIkev1PayloadVendorId(unittest.TestCase):  # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.vendor_id = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10\x11\x12\x13'
        self.large_vendor_id = bytes(range(256))

        self.test_dict_small_vendor_id = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x18'),  # 4 + 20 = 24 bytes total
            ('vendor_id', self.vendor_id),
        ])
        self.test_bytes_small_vendor_id = b''.join(self.test_dict_small_vendor_id.values())

        self.test_dict_empty_vendor_id = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE = 0x00
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x04'),
        ])
        self.test_bytes_empty_vendor_id = b''.join(self.test_dict_empty_vendor_id.values())

        self.test_dict_large_vendor_id = collections.OrderedDict([
            ('next_payload', b'\x04'),  # KEY_EXCHANGE = 0x04
            ('reserved', b'\x00'),
            ('payload_length', b'\x01\x04'),  # 4 + 256 = 260 bytes total
            ('vendor_id', self.large_vendor_id),
        ])
        self.test_bytes_large_vendor_id = b''.join(self.test_dict_large_vendor_id.values())

        # Vendor ID objects for compose tests
        self.vendor_id_small = Ikev1PayloadVendorId(vendor_id=self.vendor_id)
        self.vendor_id_small.next_payload = Ikev1PayloadType.NONE

        self.vendor_id_empty = Ikev1PayloadVendorId(vendor_id=b'')
        self.vendor_id_empty.next_payload = Ikev1PayloadType.NONE

        self.vendor_id_large = Ikev1PayloadVendorId(vendor_id=self.large_vendor_id)
        self.vendor_id_large.next_payload = Ikev1PayloadType.KEY_EXCHANGE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadVendorId.get_payload_type(), Ikev1PayloadType.VENDOR_ID)

    def test_vendor_id_with_special_characters(self):
        special_vendor_id = b'Vendor\x00\x01\x02\x03\xff\xfe\xfd'
        vendor_id = Ikev1PayloadVendorId(vendor_id=special_vendor_id)
        vendor_id.next_payload = Ikev1PayloadType.NONCE

        composed = vendor_id.compose()
        parsed: Ikev1PayloadVendorId = Ikev1PayloadVendorId.parse_exact_size(composed)

        self.assertEqual(parsed.vendor_id, special_vendor_id)
        self.assertEqual(parsed.next_payload, Ikev1PayloadType.NONCE)
        self.assertEqual(len(parsed.vendor_id), len(special_vendor_id))


class TestIkev1PayloadCertificateRequest(unittest.TestCase):
    def setUp(self):
        self.cert_encoding = 4  # X.509 certificate signature
        self.certificate_data = (
            b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
            b'\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'
        )

        self.test_dict_certreq = collections.OrderedDict([
            ('next_payload', b'\x00'),  # NONE
            ('reserved', b'\x00'),
            ('payload_length', b'\x00\x25'),  # 37 bytes (4 header + 1 cert_encoding + 32 data)
            ('cert_encoding', b'\x04'),  # X.509 certificate signature
            ('certificate_data', self.certificate_data),
        ])
        self.test_bytes_certreq = b''.join(self.test_dict_certreq.values())

        self.certreq_payload = Ikev1PayloadCertificateRequest(
            cert_encoding=self.cert_encoding,
            certificate_data=self.certificate_data
        )
        self.certreq_payload.next_payload = Ikev1PayloadType.NONE

    def test_get_payload_type(self):
        self.assertEqual(Ikev1PayloadCertificateRequest.get_payload_type(), Ikev1PayloadType.CERTIFICATE_REQUEST)

    def test_parse(self):
        parsed_certreq: Ikev1PayloadCertificateRequest = Ikev1PayloadCertificateRequest.parse_exact_size(
            self.test_bytes_certreq)
        self.assertEqual(parsed_certreq.next_payload, Ikev1PayloadType.NONE)
        self.assertEqual(parsed_certreq.cert_encoding, self.cert_encoding)
        self.assertEqual(parsed_certreq.certificate_data, self.certificate_data)

    def test_compose(self):
        composed_bytes = self.certreq_payload.compose()
        self.assertEqual(composed_bytes, self.test_bytes_certreq)

    def test_round_trip(self):
        composed_bytes = self.certreq_payload.compose()
        parsed_payload: Ikev1PayloadCertificateRequest = Ikev1PayloadCertificateRequest.parse_exact_size(
            composed_bytes)

        self.assertEqual(parsed_payload.cert_encoding, self.certreq_payload.cert_encoding)
        self.assertEqual(parsed_payload.certificate_data, self.certreq_payload.certificate_data)
        self.assertEqual(parsed_payload.next_payload, self.certreq_payload.next_payload)

    def test_error_parse_not_enough_data(self):
        incomplete_data = self.test_bytes_certreq[:-5]

        with self.assertRaises(NotEnoughData) as context_manager:
            Ikev1PayloadCertificateRequest.parse_exact_size(incomplete_data)
        self.assertEqual(context_manager.exception.bytes_needed, 5)


if __name__ == '__main__':
    unittest.main()



# SPDX-License-Identifier: MPL-2.0
# pylint: disable=too-many-lines
# -*- coding: utf-8 -*-

import abc
import collections
import itertools
import enum


import attr
import urllib3

from cryptodatahub.common.algorithm import Hash
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.types import (
    Base64Data,
    CryptoDataParamsEnumString,
    convert_base64_data,
    convert_iterable,
    convert_url,
    convert_value_to_object,
)

from cryptoparser.common.base import (
    ListParamParsable,
    ListParsable,
    Serializable,
    StringEnumCaseInsensitiveParsable,
    StringEnumParsable,
    VariantParsable,
    VariantParsableExact,
)
from cryptoparser.common.exception import InvalidType
from cryptoparser.common.field import (
    FieldParsableBase,
    FieldValueBase,
    FieldValueComponentBool,
    FieldValueComponentFloat,
    FieldValueComponentOption,
    FieldValueComponentString,
    FieldValueComponentStringBase64,
    FieldValueComponentStringEnum,
    FieldValueComponentStringEnumOption,
    FieldValueComponentTimeDelta,
    FieldValueDateTime,
    FieldValueMimeType,
    FieldValueString,
    FieldValueStringBySeparatorBase,
    FieldValueStringEnum,
    FieldValueStringEnumParams,
    FieldValueTimeDelta,
    FieldsCommaSeparated,
    FieldsJson,
    FieldsSemicolonSeparated,
    MimeTypeRegistry,
    NameValueVariantBase,
)
from cryptoparser.common.parse import ParsableBase, ParserCRLF, ParserText, ComposerText
from cryptoparser.common.utils import get_leaf_classes

from .parse import (
    HttpHeaderFieldValueComponent,
    HttpHeaderFieldValueComponentExpires,
    HttpHeaderFieldValueComponentMaxAge,
    HttpHeaderFieldValueComponentReport,
    HttpHeaderFieldValueComponentReportURI,
)


class HttpHeaderFieldValueETag(FieldValueString):
    pass


class HttpHeaderFieldValueAge(FieldValueTimeDelta):
    pass


class HttpHeaderFieldValueDate(FieldValueDateTime):
    pass


class HttpHeaderFieldValueExpires(FieldValueDateTime):
    pass


class HttpHeaderFieldValueLastModified(FieldValueDateTime):
    pass


class HttpHeaderFieldValueCacheControlMaxAge(HttpHeaderFieldValueComponentMaxAge):
    @classmethod
    def get_canonical_name(cls):
        return 'max-age'


class HttpHeaderFieldValueCacheControlSMaxAge(HttpHeaderFieldValueComponentMaxAge):
    @classmethod
    def get_canonical_name(cls):
        return 's-maxage'


class HttpHeaderFieldValueCacheControlNoCache(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'no-cache'


class HttpHeaderFieldValueCacheControlNoStore(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'no-store'


class HttpHeaderFieldValueCacheControlMustRevalidate(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'must-revalidate'


class HttpHeaderFieldValueCacheControlProxyRevalidate(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'proxy-revalidate'


class HttpHeaderFieldValueCacheControlPublic(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'public'


class HttpHeaderFieldValueCacheControlPrivate(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'private'


class HttpHeaderFieldValueCacheControlNoTransform(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'no-transform'


@attr.s
class HttpHeaderFieldValueCacheControlResponse(  # pylint: disable=too-many-instance-attributes
        FieldsCommaSeparated
):
    max_age = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueCacheControlMaxAge.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueCacheControlMaxAge)),
        default=None
    )
    s_maxage = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueCacheControlSMaxAge.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueCacheControlSMaxAge)),
        default=None
    )
    must_revalidate = attr.ib(
        converter=HttpHeaderFieldValueCacheControlMustRevalidate.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlMustRevalidate),
        default=False
    )
    proxy_revalidate = attr.ib(
        converter=HttpHeaderFieldValueCacheControlProxyRevalidate.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlProxyRevalidate),
        default=False
    )
    no_cache = attr.ib(
        converter=HttpHeaderFieldValueCacheControlNoCache.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlNoCache),
        default=False,
    )
    no_store = attr.ib(
        converter=HttpHeaderFieldValueCacheControlNoStore.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlNoStore),
        default=False,
    )
    public = attr.ib(
        converter=HttpHeaderFieldValueCacheControlPublic.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlPublic),
        default=False,
    )
    private = attr.ib(
        converter=HttpHeaderFieldValueCacheControlPrivate.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlPrivate),
        default=False,
    )
    no_transform = attr.ib(
        converter=HttpHeaderFieldValueCacheControlNoTransform.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueCacheControlNoTransform),
        default=False,
    )


class HttpHeaderFieldValueComponentIncludeSubDomains(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'includeSubDomains'


class HttpHeaderFieldValueNetworkErrorLoggingGroup(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'report_to'


class HttpHeaderFieldValueNetworkErrorLoggingMaxAge(FieldValueComponentTimeDelta):
    @classmethod
    def get_canonical_name(cls):
        return 'max_age'


class HttpHeaderFieldValueNetworkErrorLoggingIncludeSubdomains(FieldValueComponentBool):
    @classmethod
    def get_canonical_name(cls):
        return 'include_subdomains'


class HttpHeaderFieldValueNetworkErrorLoggingSuccessFraction(FieldValueComponentFloat):
    @classmethod
    def get_canonical_name(cls):
        return 'success_fraction'


class HttpHeaderFieldValueNetworkErrorLoggingFailureFraction(FieldValueComponentFloat):
    @classmethod
    def get_canonical_name(cls):
        return 'failure_fraction'


@attr.s
class HttpHeaderFieldValueNetworkErrorLogging(FieldsJson):
    report_to = attr.ib(
        converter=HttpHeaderFieldValueNetworkErrorLoggingGroup.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueNetworkErrorLoggingGroup),
    )
    max_age = attr.ib(
        converter=HttpHeaderFieldValueNetworkErrorLoggingMaxAge.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueNetworkErrorLoggingMaxAge),
    )
    include_subdomains = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueNetworkErrorLoggingIncludeSubdomains.convert),
        validator=attr.validators.optional(attr.validators.instance_of(
            HttpHeaderFieldValueNetworkErrorLoggingIncludeSubdomains
        )),
        default=None
    )
    success_fraction = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueNetworkErrorLoggingSuccessFraction.convert),
        validator=attr.validators.optional(attr.validators.instance_of(
            HttpHeaderFieldValueNetworkErrorLoggingSuccessFraction
        )),
        default=None
    )
    failure_fraction = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueNetworkErrorLoggingFailureFraction.convert),
        validator=attr.validators.optional(attr.validators.instance_of(
            HttpHeaderFieldValueNetworkErrorLoggingFailureFraction
        )),
        default=None
    )


class HttpHeaderFieldValueComponentPreload(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'preload'


@attr.s
class HttpHeaderFieldValueSTS(FieldsSemicolonSeparated):
    max_age = attr.ib(
        converter=HttpHeaderFieldValueComponentMaxAge.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge)
    )
    include_subdomains = attr.ib(
        converter=HttpHeaderFieldValueComponentIncludeSubDomains.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentIncludeSubDomains),
        default=False
    )
    preload = attr.ib(
        converter=HttpHeaderFieldValueComponentPreload.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentPreload),
        default=False
    )


@attr.s
class HttpHeaderFieldValueExpectStaple(FieldsSemicolonSeparated):
    max_age = attr.ib(
        converter=HttpHeaderFieldValueComponentMaxAge.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge)
    )
    include_subdomains = attr.ib(
        converter=HttpHeaderFieldValueComponentIncludeSubDomains.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentIncludeSubDomains),
        default=False
    )
    preload = attr.ib(
        converter=HttpHeaderFieldValueComponentPreload.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentPreload),
        default=False
    )
    report_uri = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentReportURI.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentReportURI)),
        default=None
    )


class ContentSecurityPolicyDirectiveType(StringEnumParsable, enum.Enum):
    BASE_URI = FieldValueStringEnumParams(
        code='base-uri',
    )
    BLOCK_ALL_MIXED_CONTENT = FieldValueStringEnumParams(
        code='block-all-mixed-content',
    )
    CHILD_SRC = FieldValueStringEnumParams(
        code='child-src',
    )
    CONNECT_SRC = FieldValueStringEnumParams(
        code='connect-src',
    )
    DEFAULT_SRC = FieldValueStringEnumParams(
        code='default-src',
    )
    FONT_SRC = FieldValueStringEnumParams(
        code='font-src',
    )
    FORM_ACTION = FieldValueStringEnumParams(
        code='form-action',
    )
    FRAME_ANCESTORS = FieldValueStringEnumParams(
        code='frame-ancestors',
    )
    FRAME_SRC = FieldValueStringEnumParams(
        code='frame-src',
    )
    IMG_SRC = FieldValueStringEnumParams(
        code='img-src',
    )
    MANIFEST_SRC = FieldValueStringEnumParams(
        code='manifest-src',
    )
    MEDIA_SRC = FieldValueStringEnumParams(
        code='media-src',
    )
    OBJECT_SRC = FieldValueStringEnumParams(
        code='object-src',
    )
    PLUGIN_TYPES = FieldValueStringEnumParams(
        code='plugin-types',
    )
    PREFETCH_SRC = FieldValueStringEnumParams(
        code='prefetch-src',
    )
    REFERRER = FieldValueStringEnumParams(
        code='referrer',
    )
    REPORT_SAMPLE = FieldValueStringEnumParams(
        code='report-sample',
    )
    REPORT_TO = FieldValueStringEnumParams(
        code='report-to',
    )
    REPORT_URI = FieldValueStringEnumParams(
        code='report-uri',
    )
    REQUIRE_TRUSTED_TYPES_FOR = FieldValueStringEnumParams(
        code='require-trusted-types-for',
    )
    SANDBOX = FieldValueStringEnumParams(
        code='sandbox',
    )
    SCRIPT_SRC = FieldValueStringEnumParams(
        code='script-src',
    )
    SCRIPT_SRC_ATTR = FieldValueStringEnumParams(
        code='script-src-attr',
    )
    SCRIPT_SRC_ELEM = FieldValueStringEnumParams(
        code='script-src-elem',
    )
    STYLE_SRC = FieldValueStringEnumParams(
        code='style-src',
    )
    STYLE_SRC_ATTR = FieldValueStringEnumParams(
        code='style-src-attr',
    )
    STYLE_SRC_ELEM = FieldValueStringEnumParams(
        code='style-src-elem',
    )
    TRUSTED_TYPES = FieldValueStringEnumParams(
        code='trusted-types',
    )
    UNSAFE_HASHES = FieldValueStringEnumParams(
        code='unsafe-hashes',
    )
    UPGRADE_INSECURE_REQUESTS = FieldValueStringEnumParams(
        code='upgrade-insecure-requests',
    )
    WEBRTC = FieldValueStringEnumParams(
        code='webrtc',
    )
    WORKER_SRC = FieldValueStringEnumParams(
        code='worker-src',
    )


ContentSecurityPolicySourceType = enum.Enum('ContentSecurityPolicySourceType', 'SCHEME HOST KEYWORD NONCE HASH')


class FieldHashTypeParams(CryptoDataParamsEnumString):
    pass


class StringEnumHashParsableBase(StringEnumParsable):
    @classmethod
    def from_hash_algorithm(cls, hash_algorithm):
        return cls[hash_algorithm.name]

    @property
    def hash_algorithm(self):
        return Hash[self.name]  # pylint: disable=no-member


class ContentSecurityPolicySourceHashType(StringEnumHashParsableBase, enum.Enum):
    SHA2_256 = FieldHashTypeParams(code='sha256')
    SHA2_384 = FieldHashTypeParams(code='sha384')
    SHA2_512 = FieldHashTypeParams(code='sha512')


@attr.s
class ContentSecurityPolicySourceHash(ParsableBase, Serializable):
    hash_algorithm = attr.ib(validator=attr.validators.instance_of((Hash, str)))
    hash_value = attr.ib(converter=convert_base64_data(), validator=attr.validators.instance_of(Base64Data))

    @classmethod
    def _get_hash_algorithm_enum_type(cls):
        return ContentSecurityPolicySourceHashType

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        try:
            parser.parse_parsable('hash_algorithm', cls._get_hash_algorithm_enum_type())
        except InvalidValue as e:
            raise InvalidType() from e

        parser.parse_string_until_separator_or_end('hash_value', ' ')

        return cls(parser['hash_algorithm'].hash_algorithm, parser['hash_value']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(
            self._get_hash_algorithm_enum_type().from_hash_algorithm(self.hash_algorithm).value.code
        )
        composer.compose_separator('-')
        composer.compose_string(str(self.hash_value))

        return composer.composed

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicySourceType.HASH


@attr.s
class ContentSecurityPolicySourceNonce(ParsableBase, Serializable):
    _PREFIX = 'nonce-'

    value = attr.ib(converter=convert_base64_data(), validator=attr.validators.instance_of(Base64Data))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        try:
            parser.parse_string('prefix', cls._PREFIX)
        except InvalidValue as e:
            raise InvalidType() from e

        del parser['prefix']

        parser.parse_string_until_separator_or_end('value', ' ')

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self._PREFIX)
        composer.compose_string(str(self.value))

        return composer.composed

    def _asdict(self):
        return self.value

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicySourceType.NONCE


@attr.s
class ContentSecurityPolicySourceScheme(ParsableBase, Serializable):
    value = attr.ib(validator=attr.validators.instance_of(str))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator('value', ':')
        parser.parse_separator(':')

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value)
        composer.compose_separator(':')

        return composer.composed

    def _asdict(self):
        return self.value

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicySourceType.SCHEME


@attr.s
class ContentSecurityPolicySourceHost(ParsableBase, Serializable):
    value = attr.ib(
        converter=convert_url(),
        validator=attr.validators.instance_of((str, urllib3.util.url.Url, ))
    )

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator_or_end('value', ' ')

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string(self.value)

        return composer.composed

    def _asdict(self):
        return self.value

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicySourceType.HOST


class ContentSecurityPolicySourceKeyword(StringEnumParsable, enum.Enum):
    NONE = FieldValueStringEnumParams(code='\'none\'')
    REPORT_SAMPLE = FieldValueStringEnumParams(code='\'report-sample\'')
    SELF = FieldValueStringEnumParams(code='\'self\'')
    STRICT_DYNAMIC = FieldValueStringEnumParams(code='\'strict-dynamic\'')
    UNSAFE_ALLOW_REDIRECTS = FieldValueStringEnumParams(code='\'unsafe-allow-redirects\'')
    UNSAFE_EVAL = FieldValueStringEnumParams(code='\'unsafe-eval\'')
    UNSAFE_HASHES = FieldValueStringEnumParams(code='\'unsafe-hashes\'')
    UNSAFE_INLINE = FieldValueStringEnumParams(code='\'unsafe-inline\'')
    WASM_UNSAFE_EVAL = FieldValueStringEnumParams(code='\'wasm-unsafe-eval\'')

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicySourceType.KEYWORD


@attr.s
class ContentSecurityPolicyDirectiveBase(ParsableBase, Serializable):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_type(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_parsable('type', ContentSecurityPolicyDirectiveType)
        if parser['type'] != cls.get_type():
            raise InvalidType()

        return parser

    def _compose_type(self):
        composer = ComposerText()

        composer.compose_string(self.get_type())

        return composer


class ContentSecurityPolicySerializedSource(VariantParsableExact):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (ContentSecurityPolicySourceType.KEYWORD, [ContentSecurityPolicySourceKeyword]),
            (ContentSecurityPolicySourceType.NONCE, [ContentSecurityPolicySourceNonce]),
            (ContentSecurityPolicySourceType.HASH, [ContentSecurityPolicySourceHash]),
            (ContentSecurityPolicySourceType.SCHEME, [ContentSecurityPolicySourceScheme]),
            (ContentSecurityPolicySourceType.HOST, [ContentSecurityPolicySourceHost]),
        ])


@attr.s
class ContentSecurityPolicyDirectiveSourceBase(ContentSecurityPolicyDirectiveBase):
    value = attr.ib()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_source_parser(cls):
        raise NotImplementedError()

    @value.validator
    def value_validator(self, _, value):
        self._value_validator(value)

    def _value_validator(self, value):
        source_variant_parsable = self._get_source_parser()
        acceptable_source_types = tuple(itertools.chain.from_iterable(
            source_variant_parsable._get_variants().values()  # pylint: disable=protected-access
        ))
        has_invalid_source_type = any(map(
            lambda source: not isinstance(source, acceptable_source_types),
            value
        ))
        if has_invalid_source_type:
            raise InvalidValue(value, type(self), 'value')

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_type(parsable)

        source_variant_parsable = cls._get_source_parser()

        if parser.unparsed_length:
            parser.parse_separator(' ')
            parser.parse_string_array('value', ' ', source_variant_parsable, skip_empty=True)
            directive_value = parser['value']
        else:
            raise InvalidValue(parser.unparsed, cls, 'value')

        return cls(directive_value), parser.parsed_length

    def compose(self):
        composer = self._compose_type()

        if self.value:
            composer.compose_separator(' ')
            composer.compose_string_array(self.value, ' ')

        return composer.composed

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('value', [
                collections.OrderedDict([('type', source.get_type()), ('value', source._asdict())])
                for source in self.value
            ])
        ])


class ContentSecurityPolicyDirectiveSerializedSourceListBase(ContentSecurityPolicyDirectiveSourceBase):
    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _get_source_parser(cls):
        return ContentSecurityPolicySerializedSource


class ContentSecurityPolicyDirectiveChildSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.CHILD_SRC


class ContentSecurityPolicyDirectiveConnectSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.CONNECT_SRC


class ContentSecurityPolicyDirectiveDefaultSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.DEFAULT_SRC


class ContentSecurityPolicyDirectiveFontSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.FONT_SRC


class ContentSecurityPolicyDirectiveFrameSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.FRAME_SRC


class ContentSecurityPolicyDirectiveImgSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.IMG_SRC


class ContentSecurityPolicyDirectiveManifestSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.MANIFEST_SRC


class ContentSecurityPolicyDirectiveMediaSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.MEDIA_SRC


class ContentSecurityPolicyDirectiveObjectSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.OBJECT_SRC


class ContentSecurityPolicyDirectivePrefetchSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.PREFETCH_SRC


class ContentSecurityPolicyDirectiveScriptSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.SCRIPT_SRC


class ContentSecurityPolicyDirectiveScriptSrcAttr(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.SCRIPT_SRC_ATTR


class ContentSecurityPolicyDirectiveScriptSrcElem(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.SCRIPT_SRC_ELEM


class ContentSecurityPolicyDirectiveStyleSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.STYLE_SRC


class ContentSecurityPolicyDirectiveStyleSrcAttr(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.STYLE_SRC_ATTR


class ContentSecurityPolicyDirectiveStyleSrcElem(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.STYLE_SRC_ELEM


class ContentSecurityPolicyDirectiveWorkerSrc(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.WORKER_SRC


class ContentSecurityPolicyDirectiveBaseUri(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.BASE_URI


class ContentSecurityPolicyDirectiveFormAction(ContentSecurityPolicyDirectiveSerializedSourceListBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.FORM_ACTION


class ContentSecurityPolicyFrameAncestorsSource(VariantParsableExact):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (ContentSecurityPolicySourceType.KEYWORD, [ContentSecurityPolicySourceKeyword]),
            (ContentSecurityPolicySourceType.SCHEME, [ContentSecurityPolicySourceScheme]),
            (ContentSecurityPolicySourceType.HOST, [ContentSecurityPolicySourceHost]),
        ])


class ContentSecurityPolicyDirectiveFrameAncestors(ContentSecurityPolicyDirectiveSourceBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.FRAME_ANCESTORS

    @classmethod
    def _get_source_parser(cls):
        return ContentSecurityPolicyFrameAncestorsSource

    def _value_validator(self, value):
        super()._value_validator(value)

        has_invalid_source_type = any(map(
            lambda source: (
                isinstance(source, ContentSecurityPolicySourceKeyword) and
                source not in [ContentSecurityPolicySourceKeyword.SELF, ContentSecurityPolicySourceKeyword.NONE]
            ),
            value
        ))
        if has_invalid_source_type:
            raise InvalidValue(value, type(self), 'value')


class ContentSecurityPolicyDirectiveVariant(VariantParsableExact):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (directive_class.get_type(), [directive_class, ])
            for directive_class in get_leaf_classes(ContentSecurityPolicyDirectiveBase)
        ])


@attr.s
class ContentSecurityPolicyDirectiveValueBase(ContentSecurityPolicyDirectiveBase):
    value = attr.ib()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()

    @value.validator
    def _value_validator(self, _, value):
        if not isinstance(value, self._get_value_class()):
            raise InvalidValue(value, type(self), 'value')

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_type(parsable)

        parser.parse_separator(' ')
        parser.parse_parsable('value', cls._get_value_class())

        return cls(parser['value']), parser.parsed_length

    def compose(self):
        composer = self._compose_type()

        composer.compose_separator(' ')
        composer.compose_parsable(self.value)

        return composer.composed

    def _as_markdown(self, level):
        return self._markdown_result(self.value, level)


class ContentSecurityPolicyWebRtcType(StringEnumParsable, enum.Enum):
    ALLOW = FieldValueStringEnumParams(code='\'allow\'')
    BLOCK = FieldValueStringEnumParams(code='\'block\'')


class ContentSecurityPolicyDirectiveWebrtc(ContentSecurityPolicyDirectiveValueBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.WEBRTC

    @classmethod
    def _get_value_class(cls):
        return ContentSecurityPolicyWebRtcType


class ContentSecurityPolicyReferrerPolicy(StringEnumParsable, enum.Enum):
    NO_REFERRER = FieldValueStringEnumParams(code='"no-referrer"')
    NON_WHEN_DOWNGRADE = FieldValueStringEnumParams(code='"non-when-downgrade"')
    ORIGIN = FieldValueStringEnumParams(code='"origin"')
    ORIGIN_WHEN_CROSSORIGIN = FieldValueStringEnumParams(code='"origin-when-crossorigin"')
    ORIGIN_WHEN_CROSS_ORIGIN = FieldValueStringEnumParams(code='"origin-when-cross-origin"')
    UNSAFE_URL = FieldValueStringEnumParams(code='"unsafe-url"')


class ContentSecurityPolicyDirectiveReferrer(ContentSecurityPolicyDirectiveValueBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.REFERRER

    @classmethod
    def _get_value_class(cls):
        return ContentSecurityPolicyReferrerPolicy


class ContentSecurityPolicyReportUri(FieldValueStringBySeparatorBase):
    @classmethod
    def _get_separators(cls):
        return ' "<>^`{|}'


class ContentSecurityPolicyToken(FieldValueStringBySeparatorBase):
    @classmethod
    def _get_separators(cls):
        return '"(),/:;<=>?@[\\]{} \t'


class ContentSecurityPolicyDirectiveListValueBase(ContentSecurityPolicyDirectiveBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_by_value_params(cls, parsable, value_type, value_name, value_min_length=None):
        parser = cls._parse_type(parsable)

        if parser.unparsed_length:
            parser.parse_separator(' ')
            parser.parse_string_array('value', ' ', value_type, skip_empty=True)

        if value_min_length is not None and ('value' not in parser or len(parser['value']) < value_min_length):
            raise InvalidValue(parser.unparsed, cls, value_name)

        return cls(parser['value']), parser.parsed_length

    def _compose(self, value_name):
        composer = self._compose_type()

        value = getattr(self, value_name)
        if value:
            composer.compose_separator(' ')
            composer.compose_parsable_array(value, ' ')

        return composer.composed


@attr.s
class ContentSecurityPolicyDirectiveSandbox(ContentSecurityPolicyDirectiveListValueBase):
    tokens = attr.ib(
        converter=convert_iterable(convert_value_to_object(ContentSecurityPolicyToken)),
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(ContentSecurityPolicyToken)
        )
    )

    @classmethod
    def _parse(cls, parsable):
        return cls._parse_by_value_params(parsable, ContentSecurityPolicyToken, 'tokens')

    def compose(self):
        return self._compose('tokens')

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.SANDBOX

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('tokens', self.tokens),
        ])


@attr.s
class ContentSecurityPolicyDirectivePluginTypes(ContentSecurityPolicyDirectiveListValueBase):
    mime_types = attr.ib(
        converter=convert_iterable(convert_value_to_object(FieldValueMimeType)),
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(FieldValueMimeType)
        ),
        metadata={'human_readable_name': 'MIME Types'}
    )

    @classmethod
    def _parse(cls, parsable):
        return cls._parse_by_value_params(parsable, FieldValueMimeType, 'mime_types')

    def compose(self):
        return self._compose('mime_types')

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.PLUGIN_TYPES

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('mime_types', self.mime_types),
        ])


class ContentSecurityPolicyTrustedTypeSinkGroup(StringEnumParsable, enum.Enum):
    SCRIPT = FieldValueStringEnumParams(code='\'script\'')


@attr.s
class ContentSecurityPolicyDirectiveRequireTrustedTypesFor(ContentSecurityPolicyDirectiveListValueBase):
    sink_groups = attr.ib(
        converter=convert_iterable(convert_value_to_object(ContentSecurityPolicyTrustedTypeSinkGroup)),
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(ContentSecurityPolicyTrustedTypeSinkGroup)
        ),
    )

    @classmethod
    def _parse(cls, parsable):
        return cls._parse_by_value_params(parsable, ContentSecurityPolicyTrustedTypeSinkGroup, 'sink_groups')

    def compose(self):
        return self._compose('sink_groups')

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.REQUIRE_TRUSTED_TYPES_FOR

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('sink_groups', self.sink_groups),
        ])


@attr.s
class ContentSecurityPolicyDirectiveReportUri(ContentSecurityPolicyDirectiveListValueBase):
    uri_references = attr.ib(
        converter=convert_iterable(convert_value_to_object(ContentSecurityPolicyReportUri)),
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(ContentSecurityPolicyReportUri),
        ),
        metadata={'human_readable_name': 'URI references'}
    )

    @classmethod
    def _parse(cls, parsable):
        return cls._parse_by_value_params(parsable, ContentSecurityPolicyReportUri, 'uri_references', 1)

    def compose(self):
        return self._compose('uri_references')

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.REPORT_URI

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('uri_references', self.uri_references),
        ])


@attr.s
class ContentSecurityPolicyDirectiveReportTo(ContentSecurityPolicyDirectiveBase):
    token = attr.ib(
        converter=convert_value_to_object(ContentSecurityPolicyToken),
        validator=attr.validators.instance_of(ContentSecurityPolicyToken)
    )

    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.REPORT_TO

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_type(parsable)

        parser.parse_separator(' ')
        parser.parse_parsable('token', ContentSecurityPolicyToken)

        return cls(parser['token']), parser.parsed_length

    def compose(self):
        composer = self._compose_type()

        composer.compose_separator(' ')
        composer.compose_parsable(self.token)

        return composer.composed


class ContentSecurityPolicyDirectiveNoValueBase(ContentSecurityPolicyDirectiveBase):
    @classmethod
    @abc.abstractmethod
    def get_type(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_type(parsable)

        return cls(), parser.parsed_length

    def compose(self):
        composer = self._compose_type()

        return composer.composed

    def _asdict(self):
        return collections.OrderedDict([
            ('type', self.get_type()),
            ('value', None),
        ])


class ContentSecurityPolicyDirectiveBlockAllMixedContent(ContentSecurityPolicyDirectiveNoValueBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.BLOCK_ALL_MIXED_CONTENT


class ContentSecurityPolicyDirectiveUpgradeInsecureRequests(ContentSecurityPolicyDirectiveNoValueBase):
    @classmethod
    def get_type(cls):
        return ContentSecurityPolicyDirectiveType.UPGRADE_INSECURE_REQUESTS


@attr.s
class HttpHeaderFieldValueContentSecurityPolicy(ParsableBase, Serializable):
    directives = attr.ib(
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of(ContentSecurityPolicyDirectiveBase)
        )
    )

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_array(
            'directives',
            separator=';',
            item_class=ContentSecurityPolicyDirectiveVariant,
            separator_spaces=' ',
            skip_empty=True,
        )

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string_array(self.directives, '; ')

        return composer.composed


class HttpHeaderFieldValueExpectCTComponentEnforce(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'enforce'


@attr.s
class HttpHeaderFieldValueExpectCT(FieldsCommaSeparated):
    max_age = attr.ib(
        converter=HttpHeaderFieldValueComponentMaxAge.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge)
    )
    enforce = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueExpectCTComponentEnforce.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueExpectCTComponentEnforce)),
        default=False
    )
    report_uri = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentReportURI.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentReportURI)),
        default=None
    )


class HttpHeaderFieldValueContentTypeCharset(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'charset'


class HttpHeaderFieldValueContentTypeBoundary(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'boundary'


@attr.s
class HttpHeaderFieldValueContentType(FieldsSemicolonSeparated):
    _MIME_TYPES_REQUIRE_BOUNDARY = (MimeTypeRegistry.MESSAGE, MimeTypeRegistry.MULTIPART)

    mime_type = attr.ib(
        converter=FieldValueMimeType.convert,
        validator=attr.validators.instance_of(FieldValueMimeType),
        metadata={'human_readable_name': 'MIME type'}
    )
    charset = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueContentTypeCharset.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueContentTypeCharset)),
        default=None
    )
    boundary = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueContentTypeBoundary.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueContentTypeBoundary)),
        default=None
    )

    def __attrs_post_init__(self):
        if self.mime_type.registry in self._MIME_TYPES_REQUIRE_BOUNDARY and self.boundary is None:
            raise InvalidValue(None, type(self), 'boundary')

        if self.mime_type.registry not in self._MIME_TYPES_REQUIRE_BOUNDARY and self.boundary is not None:
            raise InvalidValue(self.boundary.value, type(self), 'boundary')


class HttpHeaderXContentTypeOptions(StringEnumCaseInsensitiveParsable, enum.Enum):
    NOSNIFF = FieldValueStringEnumParams(
        code='nosniff'
    )


class HttpHeaderFieldValueXContentTypeOptions(FieldValueStringEnum):
    @classmethod
    def _get_value_type(cls):
        return HttpHeaderXContentTypeOptions


class HttpHeaderPragma(StringEnumCaseInsensitiveParsable, enum.Enum):
    NO_CACHE = FieldValueStringEnumParams(
        code='no-cache'
    )


class HttpHeaderFieldValuePragma(FieldValueStringEnum):
    @classmethod
    def _get_value_type(cls):
        return HttpHeaderPragma


class HttpHeaderFieldValuePublicKeyPinningPin(FieldValueComponentStringBase64):
    @classmethod
    def get_canonical_name(cls):
        return 'pin-sha256'


@attr.s
class HttpHeaderFieldValuePublicKeyPinning(FieldsSemicolonSeparated):
    pin_sha256 = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValuePublicKeyPinningPin.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValuePublicKeyPinningPin)),
        metadata={'human_readable_name': 'Pin (SHA-256)'}
    )
    max_age = attr.ib(
        converter=HttpHeaderFieldValueComponentMaxAge.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge),
        default=None
    )
    include_subdomains = attr.ib(
        converter=HttpHeaderFieldValueComponentIncludeSubDomains.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueComponentIncludeSubDomains),
        default=False
    )
    report_uri = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentReportURI.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentReportURI)),
        default=None
    )


class HttpHeaderFieldValueServer(FieldValueString):
    pass


class HttpHeaderFieldValueSetCookieParamDomain(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'Domain'


class HttpHeaderFieldValueSetCookieParamPath(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'Path'


class HttpHeaderFieldValueSetCookieParamSecure(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'Secure'


class HttpHeaderFieldValueSetCookieParamHttpOnly(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'HttpOnly'


class HttpHeaderSetCookieComponentSameSite(StringEnumCaseInsensitiveParsable, enum.Enum):
    STRICT = FieldValueStringEnumParams(
        code='STRICT'
    )
    LAX = FieldValueStringEnumParams(
        code='Lax'
    )
    NONE = FieldValueStringEnumParams(
        code='None'
    )


class HttpHeaderFieldValueSetCookieParamSameSite(FieldValueComponentStringEnum):
    @classmethod
    def get_canonical_name(cls):
        return 'SameSite'

    @classmethod
    def _get_value_type(cls):
        return HttpHeaderSetCookieComponentSameSite


@attr.s
class HttpHeaderFieldValueSetCookieParams(FieldsSemicolonSeparated):
    expires = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentExpires.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentExpires)),
        default=None
    )
    max_age = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentMaxAge.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge)),
        default=None
    )
    domain = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamDomain.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamDomain)),
        default=None
    )
    path = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamPath.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamPath)),
        default=None
    )
    secure = attr.ib(
        converter=HttpHeaderFieldValueSetCookieParamSecure.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamSecure),
        default=HttpHeaderFieldValueSetCookieParamSecure(False)
    )
    http_only = attr.ib(
        converter=HttpHeaderFieldValueSetCookieParamHttpOnly.convert,
        validator=attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamHttpOnly),
        default=HttpHeaderFieldValueSetCookieParamHttpOnly(False)
    )
    same_site = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamSameSite.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamSameSite)),
        default=None
    )


@attr.s
class HttpHeaderFieldValueSetCookie(FieldValueBase):  # pylint: disable=too-many-instance-attributes
    name = attr.ib(validator=attr.validators.instance_of(str))
    value = attr.ib(validator=attr.validators.instance_of(str))
    expires = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentExpires.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentExpires)),
        default=None
    )
    max_age = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentMaxAge.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentMaxAge)),
        default=None
    )
    domain = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamDomain.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamDomain)),
        default=None
    )
    path = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamPath.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamPath)),
        default=None
    )
    secure = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamSecure.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamSecure)),
        default=HttpHeaderFieldValueSetCookieParamSecure(False)
    )
    http_only = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamHttpOnly.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamHttpOnly)),
        default=HttpHeaderFieldValueSetCookieParamHttpOnly(False)
    )
    same_site = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueSetCookieParamSameSite.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueSetCookieParamSameSite)),
        default=None
    )

    @classmethod
    def _parse(cls, parsable):
        parser = ParserText(parsable)

        parser.parse_string_until_separator('name', '=')
        parser.parse_separator('=')
        parser.parse_string_until_separator_or_end('value', '; ')

        parser.parse_separator(' ', min_length=0)
        if parser.unparsed:
            parser.parse_separator(';')
        parser.parse_separator(' ', min_length=0)

        parser.parse_parsable('params', HttpHeaderFieldValueSetCookieParams)

        attributes = {
            'name': parser['name'],
            'value': parser['value'],
        }
        params = parser['params']
        attributes.update({
            name: getattr(params, name)
            for name in attr.fields_dict(type(params))
        })

        return cls(**attributes), len(parsable)

    def compose(self):
        composer = ComposerText()

        composer.compose_parsable(HttpHeaderFieldValueComponent(self.name, self.value))

        params = {}
        for name, attribute in attr.fields_dict(type(self)).items():
            value = getattr(self, name)
            if value != attribute.default and attribute.name not in ['name', 'value', ]:
                params[name] = getattr(self, name)

        if params:
            composer.compose_separator('; ')
            composer.compose_parsable(HttpHeaderFieldValueSetCookieParams(**params))

        return composer.composed


class HttpHeaderXFrameOptions(StringEnumCaseInsensitiveParsable, enum.Enum):
    DENY = FieldValueStringEnumParams(
        code='DENY'
    )
    SAMEORIGIN = FieldValueStringEnumParams(
        code='SAMEORIGIN'
    )


class HttpHeaderFieldValueXFrameOptions(FieldValueStringEnum):
    @classmethod
    def _get_value_type(cls):
        return HttpHeaderXFrameOptions


class HttpHeaderXXSSProtectionState(StringEnumParsable, enum.Enum):
    ENABLED = FieldValueStringEnumParams(
        code='1',
        human_readable_name='enabled'
    )
    DISABLED = FieldValueStringEnumParams(
        code='0',
        human_readable_name='disabled'
    )


class HttpHeaderFieldValueXXSSProtectionState(FieldValueComponentStringEnumOption):
    @classmethod
    def _get_value_type(cls):
        return HttpHeaderXXSSProtectionState


class HttpHeaderXXSSProtectionMode(StringEnumParsable, enum.Enum):
    BLOCK = FieldValueStringEnumParams(
        code='block'
    )


class HttpHeaderFieldValueXXSSProtectionMode(FieldValueComponentStringEnum):
    @classmethod
    def get_canonical_name(cls):
        return 'mode'

    @classmethod
    def _get_value_type(cls):
        return HttpHeaderXXSSProtectionMode


@attr.s
class HttpHeaderFieldValueXXSSProtection(FieldsSemicolonSeparated):
    state = attr.ib(
        converter=HttpHeaderFieldValueXXSSProtectionState,
        validator=attr.validators.instance_of(HttpHeaderFieldValueXXSSProtectionState),
    )
    mode = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueXXSSProtectionMode.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueXXSSProtectionMode)),
        default=None
    )
    report = attr.ib(
        converter=attr.converters.optional(HttpHeaderFieldValueComponentReport.convert),
        validator=attr.validators.optional(attr.validators.instance_of(HttpHeaderFieldValueComponentReport)),
        default=None
    )


class HttpHeaderReferrerPolicy(StringEnumCaseInsensitiveParsable, enum.Enum):
    NO_REFERRER = FieldValueStringEnumParams(
        code='no-referrer'
    )
    NO_REFERRER_WHEN_DOWNGRADE = FieldValueStringEnumParams(
        code='no-referrer-when-downgrade'
    )
    ORIGIN = FieldValueStringEnumParams(
        code='origin'
    )
    ORIGIN_WHEN_CROSS_ORIGIN = FieldValueStringEnumParams(
        code='origin-when-cross-origin'
    )
    SAME_ORIGIN = FieldValueStringEnumParams(
        code='same-origin'
    )
    STRICT_ORIGIN = FieldValueStringEnumParams(
        code='strict-origin'
    )
    STRICT_ORIGIN_WHEN_CROSS_ORIGIN = FieldValueStringEnumParams(
        code='strict-origin-when-cross-origin'
    )
    UNSAFE_URL = FieldValueStringEnumParams(
        code='unsafe-url'
    )


class HttpHeaderFieldValueReferrerPolicy(FieldValueStringEnum):
    @classmethod
    def _get_value_type(cls):
        return HttpHeaderReferrerPolicy


@attr.s(frozen=True)
class HttpHeaderFieldNameParams(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(str))
    normalized_name = attr.ib(validator=attr.validators.instance_of(str))

    def _as_markdown(self, level):
        return self._markdown_result(self.normalized_name, level)


class HttpHeaderFieldName(StringEnumCaseInsensitiveParsable, enum.Enum):
    AGE = HttpHeaderFieldNameParams(
        code='age',
        normalized_name='Age'
    )
    CACHE_CONTROL = HttpHeaderFieldNameParams(
        code='cache-control',
        normalized_name='Cache-Control'
    )
    CONTENT_TYPE = HttpHeaderFieldNameParams(
        code='content-type',
        normalized_name='Content-Type'
    )
    CONTENT_SECURITY_POLICY = HttpHeaderFieldNameParams(
        code='content-security-policy',
        normalized_name='Content-Security-Policy'
    )
    CONTENT_SECURITY_POLICY_REPORT_ONLY = HttpHeaderFieldNameParams(
        code='content-security-policy-report-only',
        normalized_name='Content-Security-Policy-Report-Only'
    )
    DATE = HttpHeaderFieldNameParams(
        code='date',
        normalized_name='Date'
    )
    ETAG = HttpHeaderFieldNameParams(
        code='etag',
        normalized_name='ETag'
    )
    EXPECT_CT = HttpHeaderFieldNameParams(
        code='expect-ct',
        normalized_name='Expect-CT'
    )
    EXPECT_STAPLE = HttpHeaderFieldNameParams(
        code='expect-staple',
        normalized_name='Expect-Staple'
    )
    EXPIRES = HttpHeaderFieldNameParams(
        code='expires',
        normalized_name='Expires'
    )
    LAST_MODIFIED = HttpHeaderFieldNameParams(
        code='last-modified',
        normalized_name='Last-Modified'
    )
    NETWORK_ERROR_LOGGING = HttpHeaderFieldNameParams(
        code='nel',
        normalized_name='NEL',
    )
    PRAGMA = HttpHeaderFieldNameParams(
        code='pragma',
        normalized_name='Pragma'
    )
    PUBLIC_KEY_PINNING = HttpHeaderFieldNameParams(
        code='public-key-pinning',
        normalized_name='Public-Key-Pinning'
    )
    SERVER = HttpHeaderFieldNameParams(
        code='server',
        normalized_name='Server'
    )
    SET_COOKIE = HttpHeaderFieldNameParams(
        code='set-cookie',
        normalized_name='Set-Cookie'
    )
    REFERRER_POLICY = HttpHeaderFieldNameParams(
        code='referrer-policy',
        normalized_name='Referrer-Policy'
    )
    STRICT_TRANSPORT_SECURITY = HttpHeaderFieldNameParams(
        code='strict-transport-security',
        normalized_name='Strict-Transport-Security'
    )
    X_CONTENT_SECURITY_POLICY = HttpHeaderFieldNameParams(
        code='x-content-security-policy',
        normalized_name='X-Content-Security-Policy'
    )
    X_CONTENT_TYPE_OPTIONS = HttpHeaderFieldNameParams(
        code='x-content-type-options',
        normalized_name='X-Content-Type-Options'
    )
    X_FRAME_OPTIONS = HttpHeaderFieldNameParams(
        code='x-frame-options',
        normalized_name='X-Frame-Options'
    )
    X_XSS_PROTECTION = HttpHeaderFieldNameParams(
        code='x-xss-protection',
        normalized_name='X-XSS-Protection'
    )

    @classmethod
    def from_name(cls, name):
        found_items = [
            item
            for item in cls
            if item.value.code == name.lower()
        ]

        if len(found_items) != 1:
            raise InvalidValue(name, cls, 'name')

        return found_items[0]


class HttpHeaderFieldBase(NameValueVariantBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def get_separator(cls):
        return ':'

    @classmethod
    def _compose_name_and_separator(cls, name):
        composer = cls._compose_name(name)

        composer.compose_separator(cls.get_separator())

        return composer

    def _compose_name_and_value(self, name, value):
        composer = self._compose_name_and_separator(name)

        composer.compose_separator(' ')
        composer.compose_string(value)

        return composer.composed


@attr.s
class HttpHeaderFieldParsedBase(HttpHeaderFieldBase):
    @classmethod
    @abc.abstractmethod
    def get_header_field_name(cls):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _get_value_class(cls):
        raise NotImplementedError()

    @classmethod
    def get_canonical_name(cls):
        return cls.get_header_field_name().value.code

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_name_and_separator(parsable)

        parser.parse_separator(' ', min_length=0, max_length=None)
        parser.parse_string_until_separator('value', ['\r\n', ])

        value = cls._get_value_class().parse_exact_size(parser['value'].encode('ascii'))

        return cls(value), parser.parsed_length

    def compose(self):
        return self._compose_name_and_value(
            self.get_header_field_name().value.normalized_name,
            bytes(self.value.compose()).decode('ascii')
        )

    def _asdict(self):
        return collections.OrderedDict([
            ('name', self.get_header_field_name().value.normalized_name),
            ('value', self.value)
        ])


class HttpHeaderFieldETag(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.ETAG

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueETag


class HttpHeaderFieldAge(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.AGE

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueAge


class HttpHeaderFieldContentType(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.CONTENT_TYPE

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueContentType


class HttpHeaderFieldContentSecurityPolicy(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.CONTENT_SECURITY_POLICY

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueContentSecurityPolicy


class HttpHeaderFieldXContentSecurityPolicy(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.X_CONTENT_SECURITY_POLICY

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueContentSecurityPolicy


class HttpHeaderFieldContentSecurityPolicyReportOnly(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.CONTENT_SECURITY_POLICY_REPORT_ONLY

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueContentSecurityPolicy


class HttpHeaderFieldCacheControlResponse(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.CACHE_CONTROL

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueCacheControlResponse


class HttpHeaderFieldDate(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.DATE

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueDate


class HttpHeaderFieldExpires(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.EXPIRES

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueExpires


class HttpHeaderFieldLastModified(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.LAST_MODIFIED

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueLastModified


class HttpHeaderFieldNetworkErrorLogging(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.NETWORK_ERROR_LOGGING

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueNetworkErrorLogging


class HttpHeaderFieldSTS(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.STRICT_TRANSPORT_SECURITY

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueSTS


class HttpHeaderFieldExpectCT(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.EXPECT_CT

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueExpectCT


class HttpHeaderFieldExpectStaple(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.EXPECT_STAPLE

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueExpectStaple


class HttpHeaderFieldPragma(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.PRAGMA

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValuePragma


class HttpHeaderFieldPublicKeyPinning(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.PUBLIC_KEY_PINNING

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValuePublicKeyPinning


class HttpHeaderFieldReferrerPolicy(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.REFERRER_POLICY

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueReferrerPolicy


class HttpHeaderFieldSetCookie(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.SET_COOKIE

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueSetCookie


class HttpHeaderFieldServer(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.SERVER

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueServer


class HttpHeaderFieldXContentTypeOptions(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.X_CONTENT_TYPE_OPTIONS

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueXContentTypeOptions


class HttpHeaderFieldXFrameOptions(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.X_FRAME_OPTIONS

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueXFrameOptions


class HttpHeaderFieldXXSSProtection(HttpHeaderFieldParsedBase):
    @classmethod
    def get_header_field_name(cls):
        return HttpHeaderFieldName.X_XSS_PROTECTION

    @classmethod
    def _get_value_class(cls):
        return HttpHeaderFieldValueXXSSProtection


class HttpHeaderFieldParsedVariant(VariantParsable):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (header_class.get_header_field_name(), [header_class, ])
            for header_class in get_leaf_classes(HttpHeaderFieldParsedBase)
        ])


@attr.s
class HttpHeaderFieldUnparsed(FieldParsableBase, Serializable):
    name = attr.ib(validator=attr.validators.instance_of(str))
    value = attr.ib(validator=attr.validators.instance_of(str))

    @classmethod
    def get_separator(cls):
        return ':'

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_name(parsable)
        parser.parse_separator(cls.get_separator())
        parser.parse_separator(' ', min_length=0, max_length=None)
        parser.parse_string_until_separator('value', '\r\n')

        return cls(parser['name'], parser['value']), parser.parsed_length

    def compose(self):
        composer = ComposerText()

        composer.compose_string_array([self.name, self.value], self.get_separator() + ' ')

        return composer.composed


class HttpHeaderFields(ListParsable):
    @classmethod
    def get_param(cls):
        return ListParamParsable(
            item_class=HttpHeaderFieldParsedVariant,
            fallback_class=HttpHeaderFieldUnparsed,
            separator_class=ParserCRLF
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import enum

import attr

from cryptoparser.common.base import Serializable


@attr.s(frozen=True)
class HttpVersionParams(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(str))
    name = attr.ib(validator=attr.validators.instance_of(str))

    @property
    def identifier(self):
        return self.code

    def _asdict(self):
        return self.identifier

    def _as_markdown(self, level):
        return self._markdown_result(self.name, level)


class HttpVersion(enum.Enum):
    HTTP1_0 = HttpVersionParams(code='http1_0', name='HTTP/1.0')
    HTTP1_1 = HttpVersionParams(code='http1_1', name='HTTP/1.1')



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest


from cryptodatahub.common.algorithm import Authentication, MAC, Signature
from cryptodatahub.common.exception import InvalidValue


class TestAlgortihmOIDBase(unittest.TestCase):
    def test_error_not_found(self):
        with self.assertRaisesRegex(InvalidValue, '\'1.2.3.4.5.6.7.8\' is not a valid Authentication oid value'):
            Authentication.from_oid('1.2.3.4.5.6.7.8')

    def test_from_oid(self):
        self.assertEqual(
            Authentication.from_oid(Authentication.RSA.value.oid),
            Authentication.RSA
        )


class TestAlgortihmParam(unittest.TestCase):
    def test_str(self):
        self.assertEqual(str(Signature.RSA_WITH_SHA2_224.value), 'SHA-224 with RSA Encryption')


class TestMAC(unittest.TestCase):
    def test_digest_size(self):
        self.assertEqual(MAC.SHA2_256.value.digest_size, 256)
        self.assertEqual(MAC.POLY1305.value.digest_size, 128)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc

import attr

from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData

from cryptoparser.ssh.subprotocol import (
    SshMessageBase,
    SshMessageVariantInit,
    SshMessageVariantKexDH,
    SshMessageVariantKexDHGroup,
)


@attr.s
class SshRecordBase(ParsableBase):
    HEADER_SIZE = 6

    packet = attr.ib(validator=attr.validators.instance_of(SshMessageBase))

    @classmethod
    @abc.abstractmethod
    def _get_variant_class(cls):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('packet_length', 4)
        if parser['packet_length'] > parser.unparsed_length:
            raise NotEnoughData(parser['packet_length'] - parser.unparsed_length)
        parser.parse_numeric('padding_length', 1)

        parser.parse_parsable('packet', cls._get_variant_class())

        parser.parse_raw('padding', parser['padding_length'])

        return cls(packet=parser['packet']), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        body_composer.compose_parsable(self.packet)

        payload_length = body_composer.composed_length
        padding_length = 8 - ((payload_length + 5) % 8)
        if padding_length < 4:
            padding_length += 8
        packet_length = payload_length + padding_length + 1
        for _ in range(padding_length):
            body_composer.compose_numeric(0, 1)

        header_composer = ComposerBinary()
        header_composer.compose_numeric(packet_length, 4)
        header_composer.compose_numeric(padding_length, 1)

        return header_composer.composed + body_composer.composed


class SshRecordInit(SshRecordBase):
    @classmethod
    def _get_variant_class(cls):
        return SshMessageVariantInit


class SshRecordKexDH(SshRecordBase):
    @classmethod
    def _get_variant_class(cls):
        return SshMessageVariantKexDH


class SshRecordKexDHGroup(SshRecordBase):
    @classmethod
    def _get_variant_class(cls):
        return SshMessageVariantKexDHGroup



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptodatahub.common.grade import Grade

from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.ike.version import IsakmpProtocolVersion, IsakmpVersion, IsakmpVersionFactory


class TestIsakmpProtocolVersion(unittest.TestCase):
    def setUp(self):
        self.version_bytes = b'\x11'
        self.version = IsakmpProtocolVersion(IsakmpVersion.V1, 1)

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            IsakmpProtocolVersion.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(InvalidType):
            IsakmpProtocolVersion.parse_exact_size(b'\x00')

    def test_parse(self):
        version = IsakmpProtocolVersion.parse_exact_size(self.version_bytes)
        self.assertEqual(version.major, IsakmpVersion.V1)
        self.assertEqual(version.minor, 1)

    def test_compose(self):
        self.assertEqual(self.version.compose(), self.version_bytes)

    def test_versions(self):
        version_1_0 = IsakmpProtocolVersion(IsakmpVersion.V1, 0)
        self.assertEqual(version_1_0.compose(), b'\x10')
        self.assertEqual(IsakmpProtocolVersion.parse_exact_size(b'\x10'), version_1_0)

        version_1_1 = IsakmpProtocolVersion(IsakmpVersion.V1, 1)
        self.assertEqual(version_1_1.compose(), b'\x11')
        self.assertEqual(IsakmpProtocolVersion.parse_exact_size(b'\x11'), version_1_1)

        version_2_0 = IsakmpProtocolVersion(IsakmpVersion.V2, 0)
        self.assertEqual(version_2_0.compose(), b'\x20')
        self.assertEqual(IsakmpProtocolVersion.parse_exact_size(b'\x20'), version_2_0)

    def test_identifier(self):
        self.assertEqual(IsakmpVersion.V1.identifier, "ikev1")
        self.assertEqual(IsakmpVersion.V2.identifier, "ikev2")

    def test_grade(self):
        version_v1 = IsakmpProtocolVersion(IsakmpVersion.V1, 0)
        self.assertEqual(version_v1.grade, Grade.DEPRECATED)

        version_v2 = IsakmpProtocolVersion(IsakmpVersion.V2, 0)
        self.assertEqual(version_v2.grade, Grade.SECURE)

    def test_str(self):
        version_1_1 = IsakmpProtocolVersion(IsakmpVersion.V1, 1)
        self.assertEqual(str(version_1_1), "IKEv1 (1)")

        version_2_0 = IsakmpProtocolVersion(IsakmpVersion.V2, 0)
        self.assertEqual(str(version_2_0), "IKEv2 (0)")

    def test_version(self):
        version_1_1 = IsakmpProtocolVersion(IsakmpVersion.V1, 1)
        self.assertEqual(version_1_1.version, "1.1")

        version_2_0 = IsakmpProtocolVersion(IsakmpVersion.V2, 0)
        self.assertEqual(version_2_0.version, "2.0")


class TestIsakmpVersionFactory(unittest.TestCase):
    def test_get_enum_class(self):
        self.assertEqual(IsakmpVersionFactory.get_enum_class(), IsakmpVersion)
